#include <WinTen.h>
#include <Windows.h>
#include <ClipApi.h>

void WindowBoxIn( HDC hDC, RECT * pRect );
void WindowBox( HDC hDC, RECT * pRect );
void DrawMasked( HDC hDC, HBITMAP hBmp, WORD wRow, WORD wCol );
void DrawGray( HDC hDC, HBITMAP hBmp, WORD wRow, WORD wCol );

// Dib.c
extern HGLOBAL DibRead( LPSTR szDibName );
extern LPSTR DibBits( LPBITMAPINFOHEADER lpBmp );
extern WORD far pascal wDIBColors( LPBITMAPINFOHEADER lpBmp );

HANDLE GetResources( void );

extern int PalBmpRealize( HWND hWnd, LONG hPalBmp );
extern BOOL IsForePalette( HPALETTE hPalette );

void PalBmpNew( HWND hWnd, HBITMAP hBitmap, HPALETTE hPalette );

void RegisterResource( HANDLE hRes, LPSTR szType );

static HPALETTE hPalette = 0;

//---------------------------------------------------------------------------//

void MaskRegion(HDC hdc, RECT * rct,
                       COLORREF cTransparentColor,
                       COLORREF cBackgroundColor)

{
   HDC        hdcTemp, hdcObject, hdcBack, hdcMem;
   POINT      ptSize;
   COLORREF   cColor;
   HBITMAP    bmAndObject, bmAndBack, bmBackOld, bmObjectOld,
              bmAndTemp, bmTempOld, bmAndMem, bmMemOld;
   HBRUSH     hBrush, hBrOld;

   ptSize.x = rct->right - rct->left + 1;
   ptSize.y = rct->bottom - rct->top + 1;

   hBrush      = CreateSolidBrush(cBackgroundColor);

   hdcTemp     = CreateCompatibleDC(hdc);
   hdcObject   = CreateCompatibleDC(hdc);
   hdcBack     = CreateCompatibleDC(hdc);
   hdcMem      = CreateCompatibleDC(hdc);

   bmAndTemp   = CreateCompatibleBitmap(hdc, ptSize.x, ptSize.y);
   bmAndMem    = CreateCompatibleBitmap(hdc, ptSize.x, ptSize.y);
   bmAndObject = CreateBitmap(ptSize.x, ptSize.y, 1, 1, NULL);
   bmAndBack   = CreateBitmap(ptSize.x, ptSize.y, 1, 1, NULL);

   bmTempOld   = SelectObject(hdcTemp, bmAndTemp);
   bmMemOld    = SelectObject(hdcMem, bmAndMem);
   bmBackOld   = SelectObject(hdcBack, bmAndBack);
   bmObjectOld = SelectObject(hdcObject, bmAndObject);

   hBrOld      = SelectObject(hdcMem, hBrush);

   BitBlt(hdcTemp, 0, 0, ptSize.x, ptSize.y, hdc, rct->left, rct->top, SRCCOPY);

   SetMapMode(hdcTemp, GetMapMode(hdc));

   cColor = SetBkColor(hdcTemp, cTransparentColor);

   BitBlt(hdcObject, 0, 0, ptSize.x, ptSize.y, hdcTemp, 0, 0, SRCCOPY);

   SetBkColor(hdcTemp, cColor);

   BitBlt(hdcBack, 0, 0, ptSize.x, ptSize.y, hdcObject, 0, 0, NOTSRCCOPY);
   PatBlt(hdcMem, 0,0, ptSize.x, ptSize.y, PATCOPY);
   BitBlt(hdcMem, 0, 0, ptSize.x, ptSize.y, hdcObject, 0, 0, SRCAND);
   BitBlt(hdcTemp, 0, 0, ptSize.x, ptSize.y, hdcBack, 0, 0, SRCAND);
   BitBlt(hdcMem, 0, 0, ptSize.x, ptSize.y, hdcTemp, 0, 0, SRCPAINT);
   BitBlt(hdc, rct->left, rct->top, ptSize.x, ptSize.y, hdcMem, 0, 0, SRCCOPY);

   DeleteObject(SelectObject(hdcMem, hBrOld));
   DeleteObject(SelectObject(hdcTemp, bmTempOld));
   DeleteObject(SelectObject(hdcMem, bmMemOld));
   DeleteObject(SelectObject(hdcBack, bmBackOld));
   DeleteObject(SelectObject(hdcObject, bmObjectOld));
   DeleteDC(hdcMem);
   DeleteDC(hdcBack);
   DeleteDC(hdcObject);
   DeleteDC(hdcTemp);
}

//---------------------------------------------------------------------------//

static HBITMAP CreateMemBitmap( HDC hDC, LPBITMAPINFO pBitmap )
{
   HBITMAP hBitmap = 0;

   if( hDC && pBitmap )
   {
       hBitmap =  CreateDIBitmap( hDC,
                       &pBitmap->bmiHeader,
                       CBM_INIT,
                       DibBits( &pBitmap->bmiHeader ),
                       pBitmap,
                       DIB_RGB_COLORS );
   }

   return hBitmap;
}

//---------------------------------------------------------------------------//

CLIPPER NBMPCOLORS( PARAMS ) // ( hPalette ) --> nColors
{
   HPALETTE hPalette = ( HPALETTE ) _parnl( 1 );

   if( hPalette )
   {
      _retnl( ( ( LPLOGPALETTE ) GlobalLock( ( HGLOBAL ) hPalette ) )->palNumEntries );
      GlobalUnlock( ( HGLOBAL ) hPalette );
   }
   else
      _retni( 16 );
}

//---------------------------------------------------------------------------//

#define PALVERSION 0x0300   // Windows 3.0 and later

static HPALETTE near CreateDIBPalette ( LPBITMAPINFO lpbmi )
{
   LPLOGPALETTE     lpPal;
   HANDLE           hLogPal;
   HPALETTE         hPal = 0;
   int              i, wNumColors;

   if (!lpbmi )
      return 0;

   wNumColors   = wDIBColors ( (LPBITMAPINFOHEADER) lpbmi );

   if( wNumColors > 16 )
   {
      hLogPal = GlobalAlloc ( GHND, sizeof (LOGPALETTE) +
                             sizeof (PALETTEENTRY) * wNumColors );

      if( !hLogPal )
         return 0;

      lpPal = (LPLOGPALETTE) GlobalLock ( hLogPal );

      lpPal->palVersion    = PALVERSION;
      lpPal->palNumEntries = wNumColors;

      for (i = 0;  i < wNumColors;  i++)
      {
          lpPal->palPalEntry[i].peRed   = lpbmi->bmiColors[i].rgbRed;
          lpPal->palPalEntry[i].peGreen = lpbmi->bmiColors[i].rgbGreen;
          lpPal->palPalEntry[i].peBlue  = lpbmi->bmiColors[i].rgbBlue;
          lpPal->palPalEntry[i].peFlags = 0;
      }

      hPal = CreatePalette ( lpPal );
      RegisterResource( hPal, "PAL" );

      GlobalUnlock ( hLogPal );
      GlobalFree   ( hLogPal );
   }
   return hPal;
}

//---------------------------------------------------------------------------//

HBITMAP hPalBitmapNew( HDC hDC, LPBITMAPINFO pbmi )
{
    HBITMAP  hBitmap;
    HPALETTE hPal = 0;
    BOOL     bDestroyDC = ( hDC == 0 );
    HPALETTE oldPal;

    if( bDestroyDC )
        hDC = GetDC( 0 );

    if( wDIBColors( &pbmi->bmiHeader ) > 16 )
        hPal = CreateDIBPalette ( pbmi );

    if( hPal )
    {
        oldPal = SelectPalette( hDC, hPal, FALSE );
        RealizePalette( hDC );
    }

    hBitmap = CreateMemBitmap( hDC, pbmi );
    RegisterResource( hBitmap, "BMP" );

    if( hPal )
    {
        SelectPalette( hDC, oldPal, TRUE );
        RealizePalette( hDC );
    }

    if( bDestroyDC )
        ReleaseDC( 0, hDC );

    hPalette = hPal;

    return hBitmap;
}

//---------------------------------------------------------------------------//

CLIPPER PALBMPREAD( PARAMS ) // ( hDC, cBitmapFile ) --> aBmpPal
{
   HGLOBAL hDib = DibRead( _parc( 2 ) );

   #ifdef __XPP__
      #define _stornl( x, y, z ) STORNL( x, params, y, z )
   #endif

   _reta( 2 );
   _stornl( ( LONG ) hPalBitmapNew( ( HDC ) _parnl( 1 ), (LPBITMAPINFO) GlobalLock( hDib ) ),
            -1, 1 );
   _stornl( ( LONG ) hPalette, -1, 2 );

   GlobalUnlock( hDib );
   GlobalFree( hDib );
}

//---------------------------------------------------------------------------//

CLIPPER PALBMPLOAD( PARAMS ) // ( cResourceName )
{
    HRSRC  hRes = FindResource( GetResources(),
                  IF( ISCHAR( 1 ), _parc( 1 ), MAKEINTRESOURCE( _parni( 1 ) ) ),
                  RT_BITMAP );
    HANDLE hResource = IF( hRes, LoadResource( GetResources(), hRes ), 0 );

    if( hRes )
    {
       #ifndef __FLAT__
          _reta( 2 );
          _stornl( ( LONG ) hPalBitmapNew( 0, (LPBITMAPINFO) GlobalLock( hResource ) ), -1, 1 );
          _stornl( ( LONG ) hPalette, -1, 2 );
          GlobalUnlock( hResource );
       #else
          _reta( 2 );
          _stornl( ( LONG ) hPalBitmapNew( 0, hResource ), -1, 1 );
          _stornl( ( LONG ) hPalette, -1, 2 );
       #endif
       FreeResource( hResource );
    }
    else
    {
       _reta( 2 );
       _stornl( 0, -1, 1 );
       _stornl( 0, -1, 2 );
    }
}

//---------------------------------------------------------------------------//

#ifdef __HARBOUR__
   CLIPPER PALBMPREALIZE( PARAMS ) // ( hWnd, hPalBmp )
#else
   CLIPPER PALBMPREAL( PARAMS ) // PalBmpRealize( hWnd, hPalBmp )
#endif
{
   _retni( PalBmpRealize( ( HWND ) _parnl( 1 ), _parnl( 2 ) ) );
}

//---------------------------------------------------------------------------//

void hPalBitmapDraw( HDC hDC, int wCol, int wRow, HBITMAP hBitmap, HPALETTE hPalette,
                     WORD wWidth, WORD wHeight, DWORD dwRaster, BOOL bTransparent, BOOL bVidro ) //Toninho
{
    HDC       hDcMem;
    BITMAP    bm;
    HBITMAP   hBmpOld;
    HPALETTE  oldPal;
    COLORREF  nColor;     //Toninho
    RECT      rct;

    if( ! hDC || ! hBitmap )
       return;

    if( ! dwRaster )
       dwRaster = SRCCOPY;

    _bset( ( LPBYTE ) &bm, 0, sizeof( BITMAP ) );

    GetObject( ( HGDIOBJ ) hBitmap, sizeof( BITMAP ), ( LPSTR ) &bm );

    if( wWidth && wHeight && ( bm.bmWidth != wWidth || bm.bmHeight != wHeight ) )
    {
        BITMAP  bm2;
        HBITMAP hBitmap2;
        HDC     hDcMem2;
        HPALETTE oldPal1, oldPal2;
        HBITMAP  hBmpOld2;
        int iOldMode;

        bm2 = bm;
        bm2.bmWidth      = wWidth;
        bm2.bmHeight     = wHeight;
        bm2.bmWidthBytes = ( bm2.bmWidth * bm2.bmBitsPixel + 15 ) / 16 * 2;

        hBitmap2 = CreateBitmapIndirect( &bm2 );

        hDcMem   = CreateCompatibleDC( hDC );
        hDcMem2  = CreateCompatibleDC( hDC );

        hBmpOld  = SelectObject( hDcMem, ( HGDIOBJ ) hBitmap );
        hBmpOld2 = SelectObject( hDcMem2, hBitmap2 );

        if( hPalette )
        {
            oldPal   = SelectPalette( hDC, hPalette, ! IsForePalette( hPalette ) );
            RealizePalette( hDC );
            oldPal2  = SelectPalette( hDcMem2, hPalette, FALSE );
            oldPal1  = SelectPalette( hDcMem, hPalette, FALSE );
        }

        iOldMode = SetStretchBltMode( hDC, STRETCH_DELETESCANS );

        BitBlt( hDcMem2, wRow, wCol, wWidth, wHeight,
                hDcMem, 0, 0, dwRaster );

        StretchBlt( hDC, wRow, wCol, bm2.bmWidth, bm2.bmHeight,
                    hDcMem, 0, 0, bm.bmWidth, bm.bmHeight, SRCCOPY );

        SetStretchBltMode( hDC, iOldMode );

        if( hPalette )
        {
            SelectPalette( hDcMem2, oldPal2, TRUE );
            SelectPalette( hDcMem, oldPal1, TRUE );
            SelectPalette( hDC, oldPal, TRUE );
        }

        SelectObject( hDcMem, hBmpOld );
        SelectObject( hDcMem2, hBmpOld2 );

        DeleteDC( hDcMem2 );
        DeleteObject( hBitmap2 );

    }
    else
    {
        hDcMem  = CreateCompatibleDC( hDC );
        hBmpOld = SelectObject( hDcMem, ( HGDIOBJ ) hBitmap );

        if( bVidro )                             //Toninho
          nColor = GetPixel( hDC, wRow, wCol );  //Toninho
        else                                     //Toninho
          nColor = GetSysColor( COLOR_BTNFACE ); //Toninho

        if( bTransparent )
           if( GetPixel( hDcMem, 0, 0 ) != nColor )//Toninho
            {
               rct.top = 0;
               rct.left = 0;
               rct.right = bm.bmWidth - 1;
               rct.bottom = bm.bmHeight -1;
               MaskRegion( hDcMem, &rct, GetPixel( hDcMem, 0, 0 ),
                           nColor ); //Toninho
            }

        if( hPalette )
        {
            oldPal   = SelectPalette( hDC, hPalette, ! IsForePalette( hPalette ) );
            RealizePalette( hDC );
        }

        BitBlt( hDC, wRow, wCol, bm.bmWidth, bm.bmHeight, hDcMem, 0, 0, dwRaster );

        if( hPalette )
           SelectPalette( hDC, oldPal, TRUE );

        SelectObject( hDcMem, hBmpOld );
    }
    DeleteDC( hDcMem );
}

//---------------------------------------------------------------------------//

CLIPPER PALBMPDRAW( PARAMS ) // ( hDC, row, col, hbitmap, hPalette, nHeight,
                             //   nlRaster, lTransparent )
{
   hPalBitmapDraw( ( HDC ) _parnl( 1 ), _parnl( 2 ), _parnl( 3 ),
                   ( HBITMAP ) _parnl( 4 ), ( HPALETTE ) _parnl( 5 ),
                   _parnl( 6 ), _parnl( 7 ), _parnl( 8 ), _parl( 9 ), TRUE );
}

//---------------------------------------------------------------------------//

CLIPPER NBMPHEIGHT( PARAMS ) // ( hBitmap ) --> nHeight
{
    BITMAP bm;

    GetObject( ( HGDIOBJ ) _parnl( 1 ), sizeof( BITMAP ), ( LPSTR ) &bm );

    _retnl( bm.bmHeight );
}

//---------------------------------------------------------------------------//

CLIPPER NBMPWIDTH( PARAMS ) // ( hBitmap ) --> nWidth
{
    BITMAP bm;

    GetObject( ( HGDIOBJ ) _parnl( 1 ), sizeof( BITMAP ), ( LPSTR ) &bm );

    _retnl( bm.bmWidth );
}

//---------------------------------------------------------------------------//

#ifdef __HARBOUR__
CLIPPER PALBTNPAINT( PARAMS ) // ( hWnd, hBmp1, hPal1, hBmp2, hPal2, hBmp3, hPal3,
                              //  lPressed, lAdjust, hPalBmp3, lBorder, lMenu )
#else
CLIPPER PALBTNPAIN( PARAMS )
#endif
{
    HWND hWnd      = ( HWND ) _parnl( 1 );
    HBITMAP hBmp1  = ( HBITMAP ) _parnl( 2 );
    HPALETTE hPal1 = ( HPALETTE ) _parnl( 3 );
    HBITMAP hBmp2  = ( HBITMAP ) _parnl( 4 );
    HPALETTE hPal2 = ( HPALETTE ) _parnl( 5 );
    HBITMAP hBmp3  = ( HBITMAP ) _parnl( 6 );
    HPALETTE hPal3 = ( HPALETTE ) _parnl( 7 );
    BOOL bPressed  = _parl( 8 );
    BOOL bAdjust   = _parl( 9 );
    BOOL bBorder   = _parl( 10 );
    BOOL bMenu     = _parl( 11 );
    RECT rct;
    BITMAP bm;

    HDC hDC        = GetDC( hWnd );

    GetClientRect( hWnd, &rct );

    if( hBmp3 && ! IsWindowEnabled( hWnd ) )
    {
       if ( bAdjust  )
          hPalBitmapDraw( hDC, 1, 1, hBmp3, hPal3,
                  rct.right, rct.bottom, 0, TRUE, TRUE );
       else
       {
          GetObject( ( HGDIOBJ ) hBmp3, sizeof( BITMAP ), ( LPSTR ) &bm );

          hPalBitmapDraw( hDC, ( ( rct.bottom - rct.top ) / 2 ) - ( bm.bmHeight / 2 ),
               ( ( rct.right - rct.left ) / 2 ) - ( bm.bmWidth / 2 ),
               hBmp3, hPal3, 0, 0, 0, TRUE, TRUE );

       }

       rct.bottom--;
       rct.right--;
       if( bBorder )
          WindowBox( hDC, &rct );
    }
    else
    {
       if( hBmp2 )
       {
          if( bPressed )
          {
             if ( bAdjust  )
                hPalBitmapDraw( hDC, 2, 2, hBmp2, hPal2,
                         rct.right, rct.bottom, 0, TRUE, TRUE );
             else
             {
                GetObject( ( HGDIOBJ ) hBmp2, sizeof( BITMAP ), ( LPSTR ) &bm );

                hPalBitmapDraw( hDC, ( ( rct.bottom - rct.top ) / 2 ) - ( bm.bmHeight / 2 ) + 1,
                      ( ( rct.right - rct.left ) / 2 ) - ( bm.bmWidth / 2 ) + 1,
                      hBmp2, hPal2, 0, 0, 0, TRUE, TRUE );
             }

             rct.right--;
             rct.bottom--;
             WindowBoxIn( hDC, &rct );
          }
          else
          {
             if ( bAdjust  )
                hPalBitmapDraw( hDC, 1, 1,  hBmp1, hPal1,
                         rct.right, rct.bottom, 0, TRUE, TRUE );
             else
             {
                GetObject( ( HGDIOBJ ) hBmp1, sizeof( BITMAP ), ( LPSTR ) &bm );

                hPalBitmapDraw( hDC, ( ( rct.bottom - rct.top ) / 2 ) - ( bm.bmHeight / 2 ),
                      ( ( rct.right - rct.left ) / 2 ) - ( bm.bmWidth / 2 ),
                      hBmp1, hPal1, 0, 0, 0, TRUE, TRUE );
             }

             rct.bottom--;
             rct.right--;
             if( bBorder )
                WindowBox( hDC, &rct );
          }
       }
       else
       {
          if( bPressed )
          {
             if( hBmp1 )
             {
                GetObject( ( HGDIOBJ ) hBmp1, sizeof( BITMAP ), ( LPSTR ) &bm );

                hPalBitmapDraw( hDC, ( ( rct.bottom - rct.top ) / 2 ) - ( bm.bmHeight / 2 ) + 1,
                                     ( ( rct.right - rct.left - IF( bMenu, 15, 0 ) ) / 2 ) - ( bm.bmWidth / 2 ) + 1, hBmp1, hPal1,
                       IF( bAdjust, rct.right, 0 ), IF( bAdjust, rct.bottom, 0 ), 0, TRUE, TRUE );
             }
             rct.bottom--;
             rct.right--;
             WindowBoxIn( hDC, &rct );
          }
          else
          {
             if( hBmp1 )
             {
                GetObject( ( HGDIOBJ ) hBmp1, sizeof( BITMAP ), ( LPSTR ) &bm );

                hPalBitmapDraw( hDC, ( ( rct.bottom - rct.top ) / 2 ) - ( bm.bmHeight / 2 ),
                                     ( ( rct.right - rct.left - IF( bMenu, 15, 0 ) ) / 2 ) - ( bm.bmWidth / 2 ), hBmp1, hPal1,
                       IF( bAdjust, rct.right, 0 ), IF( bAdjust, rct.bottom, 0 ), 0, TRUE, TRUE );
             }
             rct.bottom--;
             rct.right--;
             if( bBorder )
                WindowBox( hDC, &rct );
          }
       }
    }

    ReleaseDC( hWnd, hDC );
}

//---------------------------------------------------------------------------//