/*
 * Class TMultiSel Engine source code:
 * Multiple Selection for FW's TWBrowse
 *
 * Copyright 1994-2002 Hernan Diego Ceccarelli <tmultisel@cesotech.com.ar>
 *
 * Comentario: Esta clase nos da la solucion "definitiva" a la problematica
 *             de la seleccion multiple en un ambiente de browses con bases
 *             de datos. Ya no deberemos crear bases temporarias, campos des-
 *             tinados a marcas o llenar arrays temporarios, etc.,etc.,etc.
 */

#include "FiveWin.ch"
#include "wcolors.ch"


//----------------------------------------------------------------------------//

CLASS TMULTISEL

   DATA lSelectAll INIT .f.
   DATA aRecNo
   DATA oLbx
   DATA lSelect INIT .f.
   DATA nId INIT 0
   DATA lMenu INIT .t.

   METHOD  New( oLbx ) CONSTRUCTOR

   METHOD  IsSelect( nRecNo )

   METHOD  SelectAll() INLINE ::lSelectAll:= .t.,;
                              ::aRecNo:= {},;
                              ::oLbx:Refresh()

   METHOD  UnSelectAll() INLINE ::lSelectAll:= .f.,;
                                ::aRecNo:= {},;
                                ::oLbx:Refresh()

   METHOD  Reverse() INLINE ::lSelectAll:= ! ::lSelectAll,;
                            ::oLbx:Refresh()
   METHOD  Click( nRecNo, nMRow )

   MESSAGE Select( nRecNo ) METHOD _Select( nRecNo )

   METHOD  Menu( nRecNo, nMRow, nMCol )

ENDCLASS

//----------------------------------------------------------------------------//

METHOD New( oLbx ) CLASS TMULTISEL
 LOCAL nClrPaneSel:= GetSysColor( COLOR_HIGHLIGHT )
 LOCAL nClrTextSel:= GetSysColor( COLOR_HIGHLIGHTTEXT )
 LOCAL nClrPane   := GetSysColor( COLOR_WINDOW )
 LOCAL nClrText   := GetSysColor( COLOR_WINDOWTEXT )
 LOCAL bLClickedOld, bRClickedOld, bLineOld, bKeyDownOld, cAlias

    ::aRecNo:= {}
    ::oLbx:= oLbx
    cAlias:= oLbx:cAlias

    If ( bLClickedOld:= oLbx:bLClicked ) != Nil
       oLbx:bLClicked:= {|nMRow,nMCol,nFlags| ;
                         ::Click( (cAlias)->(RecNo()), nMRow ),;
                         Eval( bLClickedOld, nMRow, nMCol, nFlags ) }
    Else
       oLbx:bLClicked:= {|nMRow| ::Click( (cAlias)->(RecNo()), nMRow ) }
    EndIf


    If ( bRClickedOld:= oLbx:bRClicked ) != Nil
       oLbx:bRClicked:= {|nMRow,nMCol,nFlags| ;
                         ::Menu( (cAlias)->(RecNo()), nMRow, nMCol ),;
                         Eval( bRClickedOld, nMRow, nMCol, nFlags ) }
    Else
       oLbx:bRClicked:= {|nMRow,nMCol| ::Menu( (cAlias)->(RecNo()), nMRow, nMCol ) }
    EndIf


    If ( bKeyDownOld:= oLbx:bKeyDown ) != Nil
       oLbx:bKeyDown:= {|nKey,nFlags| ;
                        If( nKey==VK_SPACE,;
                            ::Select((cAlias)->(RecNo())), ),;
                        Eval( bKeyDownOld, nKey, nFlags ) }
    Else
       oLbx:bKeyDown:= {|nKey| ;
                        If( nKey== VK_SPACE,;
                            ::Select((cAlias)->(RecNo())), ) }
    EndIf


    bLineOld:= oLbx:bLine
    oLbx:bLine:= {|| ::IsSelect( (cAlias)->(RecNo()), .t. ),;
                     Eval( bLineOld ) }

    oLbx:lDrawHeaders:= .f.

    oLbx:nLineStyle:= 0
    oLbx:lMChange:= .f.

    oLbx:nClrText:= {|| If( ::lSelect, nClrTextSel, nClrText ) }
    oLbx:nClrPane:= {|| If( ::lSelect, nClrPaneSel, nClrPane ) }
    // Celda Activa Cuando es Foused
    oLbx:nClrForeFocus := oLbx:nClrText
    oLbx:nClrBackFocus := oLbx:nClrPane

    oLbx:nClrNFFore := oLbx:nClrText
    oLbx:nClrNFBack := oLbx:nClrPane


return Self

//----------------------------------------------------------------------------//

METHOD IsSelect( nRecNo, lAssign ) CLASS TMULTISEL
 LOCAL lSelect:= .f.

    DEFAULT lAssign:= .f.

    ::nId:= AScan( ::aRecNo, nRecNo )

    If ::lSelectAll  // Si estan seleccionados todos
       If ::nId > 0
          lSelect:= .f.
       Else
          lSelect:= .t.
       EndIf
    Else             // Si NO estan seleccionados todos
       If ::nId > 0
          lSelect:= .t.
       EndIf
    EndIf

    If lAssign
       ::lSelect:= lSelect
    EndIf

return lSelect

//----------------------------------------------------------------------------//

METHOD Click( nRecNo, nMRow ) CLASS TMULTISEL
 LOCAL nRow:= ::oLbx:nWRow( nMRow )

    If nRow < 1 .or. nRow > ::oLbx:nRowCount()
       return nil
    EndIf
    ::Select( nRecNo )
return Nil

//----------------------------------------------------------------------------//

METHOD _Select( nRecNo ) CLASS TMULTISEL

    If nRecNo > 0
       If ::nId > 0
          ADel( ::aRecNo, ::nId )
          ASize( ::aRecNo, Len( ::aRecNo )-1 )
       Else
          Aadd( ::aRecNo, nRecNo )
       EndIf
       ::oLbx:Refresh(.f.)
    EndIf
return Nil

//----------------------------------------------------------------------------//

METHOD Menu( nRecNo, nMRow, nMCol )
 LOCAL oMenu

    If ::lMenu

       MENU oMenu POPUP

             If ::lSelect
                MENUITEM "&Desmarcar Item" ACTION ::Select( nRecNo )
             Else
                MENUITEM "&Marcar Item" ACTION ::Select( nRecNo )
             EndIf

             MENUITEM "Seleccionar &todo" ACTION ::SelectAll()

             MENUITEM "DeSele&ccionar todo" ACTION ::UnSelectAll()

             SEPARATOR

             MENUITEM "Revertir Seleccin" ACTION ::Reverse()

       ENDMENU

       ACTIVATE POPUP oMenu AT nMRow, nMCol OF ::oLbx

    EndIf

return 0

//----------------------------------------------------------------------------//

